
/*! \file
* Copyright (c) 2008
* Project : Base
* File    : Font Source (Independent)
* Author  : Dave Reed
*
* Loads fonts created with the Angelcode Bitmap Font Generator
*/

/*-Includes------------------------------------------------------------------*/

#include "Main.h"
#include "TinyXML.h"
#include "Base/Font.h"
#include "Base/TiXmlUtil.h"

Font::Font(char *szFontName, float fExtraSpacing, float fExtraSpaceWidth)
{
	// Init
	m_iNumChars   = 0;
	m_fLineHeight = 0;
	m_fBaseHeight = 0;

	for(int i=0; i<256; i++)
	{
		m_axChars[i].m_bValid    = false;
		m_axChars[i].m_fxAdvance = 0.0f;
		m_axChars[i].m_fxOffset  = 0.0f;
		m_axChars[i].m_fyOffset  = 0.0f;
	}

	m_apxFontTextures[0] = NULL;
	m_apxFontTextures[1] = NULL;

	// Open the XML font definition
	char szFileName[MAX_FILENAME_LENGTH];

	sprintf(szFileName, "data\\fonts\\%s.fnt", szFontName);

	TiXmlDocument XMLDoc( szFileName );	
	if (!XMLDoc.LoadFile())
	{
		return;
	}

	// Load the textures
	sprintf(szFileName, "fonts\\%s_00.png", szFontName);
	m_apxFontTextures[0] = Texture::LoadTexture(szFileName, true, 1);
	
	// Background texture, if we have one	
	sprintf(szFileName, "fonts\\%s_00_bg.png", szFontName);
	m_apxFontTextures[1] = Texture::LoadTexture(szFileName, true, 1);

	float fTexWidth, fTexHeight;
	
	if (m_apxFontTextures[0]==NULL)
	{
		// No main texture, can't continue
		return;
	}
	else
	{
		fTexWidth  = (float)m_apxFontTextures[0]->GetWidth();
		fTexHeight = (float)m_apxFontTextures[0]->GetHeight();
	}

	// Parse the XML
	TiXmlElement* pRootElt = XMLDoc.FirstChildElement( "font" );
	if (pRootElt)
	{
		TiXmlElement* pCommonElt = pRootElt->FirstChildElement( "common" );
		if(pCommonElt)
		{
			 m_fBaseHeight = TiXmlUtil::GetFloat(pCommonElt, "base", 0);			 
			 m_fLineHeight = TiXmlUtil::GetFloat(pCommonElt, "lineHeight", 0);
		}

		TiXmlElement* pCharsElt = pRootElt->FirstChildElement( "chars" );
		if(pCharsElt)
		{
			m_iNumChars = TiXmlUtil::GetInt(pCharsElt, "count", 0);

			TiXmlElement *pCharElt = pCharsElt->FirstChildElement ( "char" );
			
			while (pCharElt)
			{
			
				int   iChar = TiXmlUtil::GetInt(pCharElt, "id", -1);

				if (iChar>=0 && iChar<256)
				{
					float fX = TiXmlUtil::GetFloat(pCharElt, "x",	   0.0f);
					float fY = TiXmlUtil::GetFloat(pCharElt, "y",	   0.0f);
					float fW = TiXmlUtil::GetFloat(pCharElt, "width",  0.0f);
					float fH = TiXmlUtil::GetFloat(pCharElt, "height", 0.0f);

					m_axChars[iChar].m_afU[0] = fX		/ fTexWidth;
					m_axChars[iChar].m_afU[1] = (fX+fW) / fTexWidth;
					m_axChars[iChar].m_afV[0] = fY		/ fTexHeight;
					m_axChars[iChar].m_afV[1] = (fY+fH) / fTexHeight;

					m_axChars[iChar].m_fWidth    = fW;
					m_axChars[iChar].m_fHeight   = fH;
					m_axChars[iChar].m_fxOffset  = TiXmlUtil::GetFloat(pCharElt, "xoffset",  0.0f);
					m_axChars[iChar].m_fyOffset  = TiXmlUtil::GetFloat(pCharElt, "yoffset",  0.0f);
					m_axChars[iChar].m_fxAdvance = fExtraSpacing + TiXmlUtil::GetFloat(pCharElt, "xadvance", 0.0f);
					m_axChars[iChar].m_bValid = true;

					if (iChar==32 && fExtraSpaceWidth>0.0f) m_axChars[iChar].m_fxAdvance+=fExtraSpaceWidth;
				}

				pCharElt = pCharElt->NextSiblingElement( "char" );
			}
		}
	}
}


Font::~Font()
{
	for(int i=0; i<2; i++)
	{
		if (m_apxFontTextures[i])
		{
			Texture::FlushTexture(m_apxFontTextures[i]);
			m_apxFontTextures[i] = NULL;
		}
	}
}



float Font::GetStringLength(char *szString, float fScale)
{
	int iLen = (int)strlen(szString);
	if (iLen > MAX_FILENAME_LENGTH) iLen = MAX_FILENAME_LENGTH;

	float fLen=0.0f;
	for(int i=0; i<iLen; i++)
	{
		int iChar = szString[i];

		if (m_axChars[iChar].m_bValid)
		{
			fLen += m_axChars[iChar].m_fxAdvance;			
		}
	}

	return fLen*fScale;
}

void Font::PrintString(char *szString, float fX, float fY, int iFlags, uint32 uColor, float fScale)
{
	static BlitRect2D s_xRects[MAX_FILENAME_LENGTH];

	int iLen = (int)strlen(szString);
	if (iLen > MAX_FILENAME_LENGTH) iLen = MAX_FILENAME_LENGTH;

	float fStartX = fX;
	float fStartY = fY;

	if (iFlags & ALIGN_BOTTOM) fStartY -= (m_fLineHeight * fScale);
	if (iFlags & ALIGN_MIDDLE) fStartY -= (m_fLineHeight * fScale * 0.5f);

	float fWidth = GetStringLength(szString, fScale);
	if (iFlags & ALIGN_RIGHT)  fStartX -= fWidth;
	if (iFlags & ALIGN_CENTRE) fStartX -= (fWidth * 0.5f);
	

	for(int i=1; i>=0; i--)
	{
		float fCurrX = fStartX;
		float fCurrY = fStartY;

		if (m_apxFontTextures[i])
		{
			int iCount=0;

			for(int j=0; j<iLen; j++)
			{
				int iChar = szString[j];

				if (m_axChars[iChar].m_bValid)
				{
					FontChar   *pxChar = &m_axChars[iChar];
					BlitRect2D *pxRect = &s_xRects[iCount];

					pxRect->fX[0] = fCurrX + pxChar->m_fxOffset * fScale;
					pxRect->fY[0] = fCurrY + pxChar->m_fyOffset * fScale;
					pxRect->fX[1] = pxRect->fX[0] + pxChar->m_fWidth * fScale;
					pxRect->fY[1] = pxRect->fY[0] + pxChar->m_fHeight * fScale;

					pxRect->fU[0] = pxChar->m_afU[0];
					pxRect->fU[1] = pxChar->m_afU[1];
					pxRect->fV[0] = pxChar->m_afV[0];
					pxRect->fV[1] = pxChar->m_afV[1];

					pxRect->uARGB = uColor;

					fCurrX += pxChar->m_fxAdvance * fScale;
					iCount++;
				}
			}

			Blitter2D::BlitRects(iCount, s_xRects, m_apxFontTextures[i]);
		}
	}
}

