/*! \file
 * Copyright (c) 2008
 * Project : Base
 * File    : Math Source (Independent)
 * Author  : Dave Reed
 */

/*-Project-Includes----------------------------------------------------------*/

#include "Main.h"
#include "Base/Math.h"

/*-Constants-----------------------------------------------------------------*/

const float MATH_PI = 3.14159265f;
const float MATH_RADIANS_TO_DEGREES = (180.0f / MATH_PI);
const float MATH_DEGREES_TO_RADIANS = (MATH_PI / 180.0f);

/*-Class-Public-Functions----------------------------------------------------*/

/*!
 * \brief  Clamps a float to a given range
 * \param  fValue The value to be clamped
 * \param  fMin Minimum value
 * \param  fMax Maximum Value
 * \return The clamped value
 */
float Math::Clamp(const float fValue, const float fMin, const float fMax)
{
	float fOutput = fValue;
	if ( fValue > fMax ) fOutput = fMax;
	if ( fValue < fMin ) fOutput = fMin;
	return fOutput;
}

/*!
 * \brief  Linearly interpolates between two values
 * \param  fSrc Source value
 * \param  fDest Destination value
 * \param  fLerp Lerp value (0-1)
 * \return The interpolated value
 */
float Math::Lerp(const float fSrc, const float fDest, float fLerp)
{
	float fClampedLerp = Clamp( fLerp, 0.0f, 1.0f );
	return fSrc * (1.0f-fClampedLerp) + fDest * fClampedLerp;
}

/*!
 * \brief  Calculates sin(fRadians)
 * \param  fRadians Input angle in radians
 * \return sin(fRadians)
 */
float Math::Sin( const float fRadians )
{
	return sin(fRadians);
}

/*!
 * \brief  Calculates cos(fRadians)
 * \param  fRadians Input angle in radians
 * \return cin(fRadians)
 */
float Math::Cos( const float fRadians )
{
	return cos(fRadians);
}

/*!
 * \brief  Calculates tan(fRadians)
 * \param  fRadians Input angle in radians
 * \return tan(fRadians)
 */
float Math::Tan( const float fRadians )
{
	return tan(fRadians);
}

/*!
 * \brief  Calculates arcsin(fRadians)
 * \param  fSin Input value
 * \return arcsin(fRadians)
 */
float Math::ASin( const float fSin )
{
	return asin(fSin);
}

/*!
 * \brief  Calculates arccos(fRadians)
 * \param  fCos Input value
 * \return arccos(fRadians)
 */
float Math::ACos( const float fCos )
{
	return acos(fCos);
}

/*!
 * \brief  Calculates arctan(fRadians)
 * \param  fTan Input value
 * \return arctan(fRadians)
 */
float Math::ATan( const float fTan )
{
	return atan(fTan);
}

/*!
 * \brief  Converts degrees to radians
 * \param  fDegrees Angle in degrees
 * \return Angle in radians
 */
float Math::DegToRad( const float fDegrees )
{
	return fDegrees * MATH_DEGREES_TO_RADIANS;
}

/*!
 * \brief  Converts radians to degrees
 * \param  fRadians Angle in radians
 * \return Angle in degrees
 */
float Math::RadToDeg( const float fRadians )
{
	return fRadians * MATH_RADIANS_TO_DEGREES;
}

/*!
 * \brief  Square Root
 * \param  fInput Input Value
 * \return Square root of input value
 */
float Math::Sqrt( const float fInput )
{
	return sqrt(fInput);
}

/*!
 * \brief  Floating-point abs()
 * \param  fInput Input Value
 * \return Absolute value of input value
 */
float Math::FAbs( const float fInput )
{
	return (fInput<0.0f) ? -fInput : fInput;
}

/*!
 * \brief  Returns Pi (to 8 decimal places)
 * \return Pi
 */
float Math::Pi( void )
{
	return MATH_PI;
}
