/*! \file
 * Copyright (c) 2008
 * Project : Base
 * File    : Matrix Source (Independent)
 * Author  : Dave Reed
 */

/*-Project-Includes----------------------------------------------------------*/

#include "Main.h"
#include "Base/Matrix.h"
#include "Base/Math.h"
#include "Base/Vector.h"

/*-Class-Public-Functions----------------------------------------------------*/

/* Constructors */

/*!
 * \brief  Default constructor (Inits to identity matrix)
 */
Matrix::Matrix( void )
{
	SetIdentity();
}

/*!
 * \brief  'Fast constructor' (if we know it doesn't need initialising)
 * \param  bSetIdentity true to init the matrix to the identity matrix, false to leave it uninitialised
 */
Matrix::Matrix( const bool bSetIdentity )
{
	if (bSetIdentity)
	{
		SetIdentity();
	}
}

/*!
 * \brief  Copy constructor
 * \param  xMtx Matrix to copy from
 */
Matrix::Matrix( const Matrix &xMtx )
{
	*this = xMtx;
}

/*!
 * \brief  Assignment Operator
 * \param  xV Matrix to copy
 */
Matrix& Matrix::operator = ( const Matrix &xV )
{
	for(int i=0; i<4; i++)
	{
		for(int j=0; j<4; j++)
		{
			m_afElts[i][j] = xV.m_afElts[i][j];
		}
	}

	return *this;
}

/*!
 * \brief  Indexing Operator
 * \param  iElement Row to return
 * \return Matrix row, as a vector (Note: W component may need setting to 1)
 */
Vector &Matrix::operator [] ( const int iElement ) const
{
	Vector *pfVec = (Vector*)&m_afElts[iElement][0];
	return (Vector&)(*pfVec);
}

/*!
 * \brief Sets matrix to the identity matrix
 */
void Matrix::SetIdentity()
{
	m_afElts[0][0] = 1.0f;	m_afElts[0][1] = 0.0f;	m_afElts[0][2] = 0.0f;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = 0.0f;	m_afElts[1][1] = 1.0f;	m_afElts[1][2] = 0.0f;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] = 0.0f;	m_afElts[2][1] = 0.0f;	m_afElts[2][2] = 1.0f;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Sets matrix to a translation of xV
 * \param xV Translation vector
 */
void Matrix::SetTranslation( const Vector &xV )
{
	m_afElts[0][0] = 1.0f;	m_afElts[0][1] = 0.0f;	m_afElts[0][2] = 0.0f;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = 0.0f;	m_afElts[1][1] = 1.0f;	m_afElts[1][2] = 0.0f;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] = 0.0f;	m_afElts[2][1] = 0.0f;	m_afElts[2][2] = 1.0f;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = xV.m_fX;
	m_afElts[3][1] = xV.m_fY;
	m_afElts[3][2] = xV.m_fZ;
	m_afElts[3][3] = 1.0f;
}

/*!
 * \brief Sets matrix to a scale of (fX,fY,fZ)
 * \param fX X-axis scale factor
 * \param fY Y-axis scale factor
 * \param fZ Z-axis scale factor
 */
void Matrix::SetScale( const float fX, const float fY, const float fZ )
{
	m_afElts[0][0] = fX;	m_afElts[0][1] = 0.0f;	m_afElts[0][2] = 0.0f;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = 0.0f;	m_afElts[1][1] = fY;	m_afElts[1][2] = 0.0f;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] = 0.0f;	m_afElts[2][1] = 0.0f;	m_afElts[2][2] = fZ;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Sets matrix to a rotation about the X axis
 * \param fRadians Angle in radians
 */
void Matrix::SetRotationX( const float fRadians )
{
	float fSin = Math::Sin(fRadians);
	float fCos = Math::Cos(fRadians);

	m_afElts[0][0] = 1.0f;	m_afElts[0][1] = 0.0f;	m_afElts[0][2] = 0.0f;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = 0.0f;	m_afElts[1][1] = fCos;	m_afElts[1][2] =-fSin;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] = 0.0f;	m_afElts[2][1] = fSin;	m_afElts[2][2] = fCos;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Sets matrix to a rotation about the Y axis
 * \param fRadians Angle in radians
 */
void Matrix::SetRotationY( const float fRadians )
{
	float fSin = Math::Sin(fRadians);
	float fCos = Math::Cos(fRadians);

	m_afElts[0][0] = fCos;	m_afElts[0][1] = 0.0f;	m_afElts[0][2] = fSin;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = 0.0f;	m_afElts[1][1] = 1.0f;	m_afElts[1][2] = 0.0f;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] =-fSin;	m_afElts[2][1] = 0.0f;	m_afElts[2][2] = fCos;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Sets matrix to a rotation about the Y axis
 * \param fRadians Angle in radians
 */
void Matrix::SetRotationZ( const float fRadians )
{
	float fSin = Math::Sin(fRadians);
	float fCos = Math::Cos(fRadians);

	m_afElts[0][0] = fCos;	m_afElts[0][1] =-fSin;	m_afElts[0][2] = 0.0f;	m_afElts[0][3] = 0.0f;  
	m_afElts[1][0] = fSin;	m_afElts[1][1] = fCos;	m_afElts[1][2] = 0.0f;	m_afElts[1][3] = 0.0f;  
	m_afElts[2][0] = 0.0f;	m_afElts[2][1] = 0.0f;	m_afElts[2][2] = 1.0f;	m_afElts[2][3] = 0.0f;  
	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Sets matrix to a rotation about an arbitrary axis
 * \param xAxisIn Axis vector (doesn't need to be unit length)
 * \param fRadians Angle in radians
 */
void Matrix::SetRotation( const Vector xAxisIn, const float fRadians )
{
	// From http://www.gamedev.net/reference/articles/article1199.asp, originally from Graphics Gems
	float fSin = Math::Sin(fRadians);
	float fCos = Math::Cos(fRadians);
	float fT = 1-fCos;

	Vector xAxis = xAxisIn;
	xAxis.Normalize();

	m_afElts[0][0] = (fT * xAxis.m_fX * xAxis.m_fX) + fCos;
	m_afElts[0][1] = (fT * xAxis.m_fX * xAxis.m_fY) + (fSin * xAxis.m_fZ);
	m_afElts[0][2] = (fT * xAxis.m_fX * xAxis.m_fZ) - (fSin * xAxis.m_fY);
	m_afElts[0][3] = 0.0f;  

	m_afElts[1][0] = (fT * xAxis.m_fY * xAxis.m_fX) - (fSin * xAxis.m_fZ);
	m_afElts[1][1] = (fT * xAxis.m_fY * xAxis.m_fY) + fCos;
	m_afElts[1][2] = (fT * xAxis.m_fY * xAxis.m_fZ) + (fSin * xAxis.m_fX);
	m_afElts[1][3] = 0.0f;  

	m_afElts[2][0] = (fT * xAxis.m_fZ * xAxis.m_fX) + (fSin * xAxis.m_fY);
	m_afElts[2][1] = (fT * xAxis.m_fZ * xAxis.m_fY) - (fSin * xAxis.m_fX);
	m_afElts[2][2] = (fT * xAxis.m_fZ * xAxis.m_fZ) + fCos;
	m_afElts[2][3] = 0.0f;  

	m_afElts[3][0] = 0.0f;	m_afElts[3][1] = 0.0f;	m_afElts[3][2] = 0.0f;	m_afElts[3][3] = 1.0f;  
}

/*!
 * \brief Transposes the matrix (the entire 4x4)
 */
void Matrix::Transpose( void )
{
	Matrix xOld = *this;
	for(int j=0; j<4; j++)
		for(int i=0; i<4; i++)
		{
			m_afElts[i][j] = xOld.m_afElts[j][i];
		}
}

/*!
 * \brief Transposes the upper left 3x3 elements of the matrix
 */
void Matrix::Transpose3x3( void )
{
	Matrix xOld = *this;
	for(int j=0; j<3; j++)
		for(int i=0; i<3; i++)
		{
			m_afElts[i][j] = xOld.m_afElts[j][i];
		}
}

/*!
 * \brief Simple+Fast matrix inverse for matrices with no scale/shear/projection components
 */
void Matrix::InverseUnscaled( void )
{
	Transpose3x3();
	Vector xTranslation = GetTranslationVector();
	xTranslation = PreTransform3x3(xTranslation);
	SetTranslationVector( - xTranslation );
}

/*!
 * \brief Gets a 3x3 submatrix of a 4x4 matrix - For determinant/inverse calculation
 * \param xDst Destination matrix
 * \param iRowSkip Row to skip
 * \param iColSkip Column to skip
 */
void Matrix::GetSubMatrix(Matrix &xDst, int iRowSkip, int iColSkip) const
{
	for( int i = 0; i < 3; i ++ ) 
		for( int j = 0; j < 3; j ++ )
		{
			// Map 3x3 destination element to 4x4 source element, skipping row iSkip and column jSkip
			xDst.m_afElts[i][j] = m_afElts[i + ((i >= iRowSkip)?1:0)][j + ((j >= iColSkip)?1:0)];
		}
}

/*!
 * \brief Calculates the determinant of the matrix (4x4)
 */
float Matrix::Determinant(void) const
{
	float fDet;
	float fResult = 0.0f;
	float fSign = 1.0f;

	Matrix xSubMtx(false);

	for ( int i = 0; i < 4; i++, fSign *= -1.0f )
	{
		GetSubMatrix(xSubMtx, 0, i);
		fDet     = xSubMtx.Determinant3x3();
		fResult += m_afElts[0][i] * fDet * fSign;
	}

	return fResult;
}

/*!
 * \brief Calculates the determinant of the matrix (3x3)
 */
float Matrix::Determinant3x3(void) const
{
	return	  m_afElts[0][0] * ( m_afElts[1][1] * m_afElts[2][2] - m_afElts[2][1] * m_afElts[1][2] )
		- m_afElts[0][1] * ( m_afElts[1][0] * m_afElts[2][2] - m_afElts[2][0] * m_afElts[1][2] )
		+ m_afElts[0][2] * ( m_afElts[1][0] * m_afElts[2][1] - m_afElts[2][0] * m_afElts[1][1] );
}

/*!
 * \brief Computes a proper inverse of an arbitrary 4x4 matrix, if possible
 * \return True if the inverse was successful
 */
bool Matrix::Inverse( void )
{
	float fDet = Determinant();

	if ( Math::FAbs( fDet ) < 0.0005f )
	{
		// Not inverseable
		return false;
	}
	else
	{
		int    i, j;
		float  fSign;
		Matrix xSubMtx(false);
		Matrix xOriginalMtx = *this;

		for ( i = 0; i < 4; i++ )
		{
			for ( j = 0; j < 4; j++ )
			{
				fSign = 1.0f - ((i + j) & 1) * 2.0f;
				xOriginalMtx.GetSubMatrix( xSubMtx, i, j );
				m_afElts[j][i] = ( xSubMtx.Determinant3x3() * fSign ) / fDet;
			}
		}

		return true;
	}
}

/* Decomposition */

/*!
 * \brief Extracts the right-vector from the matrix
 * \return Right vector (Row 0 of the matrix, with 1.0 put in the W component)
 */
Vector Matrix::GetRightVector( void ) const
{
	return Vector(m_afElts[0][0], m_afElts[0][1], m_afElts[0][2], 1.0f);
}

void Matrix::SetRightVector( Vector &v )
{
	m_afElts[0][0] = v.m_fX;
	m_afElts[0][1] = v.m_fY;
	m_afElts[0][2] = v.m_fZ;
}

/*!
 * \brief Extracts the up-vector from the matrix
 * \return Right vector (Row 1 of the matrix, with 1.0 put in the W component)
 */
Vector Matrix::GetUpVector( void ) const
{
	return Vector(m_afElts[1][0], m_afElts[1][1], m_afElts[1][2], 1.0f);
}

void Matrix::SetUpVector( Vector &v )
{
	m_afElts[1][0] = v.m_fX;
	m_afElts[1][1] = v.m_fY;
	m_afElts[1][2] = v.m_fZ;
}

/*!
 * \brief Extracts the forward-vector from the matrix
 * \return Right vector (Row 2 of the matrix, with 1.0 put in the W component)
 */
Vector Matrix::GetForwardVector( void ) const
{
	return Vector(m_afElts[2][0], m_afElts[2][1], m_afElts[2][2], 1.0f);
}

void Matrix::SetForwardVector( Vector &v )
{
	m_afElts[2][0] = v.m_fX;
	m_afElts[2][1] = v.m_fY;
	m_afElts[2][2] = v.m_fZ;	
}


/*!
 * \brief Extracts the translation vector from the matrix
 * \return Translation vector (Row 3 of the matrix, with 1.0 put in the W component)
 */
Vector Matrix::GetTranslationVector( void ) const
{
	return Vector(m_afElts[3][0], m_afElts[3][1], m_afElts[3][2], 1.0f);
}

/*!
 * \brief Sets the translation vector of the matrix (first 3 elements of row 3)
 * \param xV New translation vector
 */
void Matrix::SetTranslationVector( const Vector xV )
{
	m_afElts[3][0] = xV.m_fX;
	m_afElts[3][1] = xV.m_fY;
	m_afElts[3][2] = xV.m_fZ;
}

/*!
 * \brief  Performs a 4x4 matrix multiply ( this * xMat )
 * \param  xMat Matrix to multiply by
 * \return Result of the multiply operation
 */
Matrix Matrix::operator*  ( const Matrix &xMat ) const
{
	return Multiply(xMat);
}

/*!
 * \brief  Performs a 4x4 matrix multiply ( this * xMat )
 * \param  xMat Matrix to multiply by
 * \return Result of the multiply operation
 */
Matrix Matrix::Multiply   ( const Matrix &xMat ) const
{
	Matrix xResult = Matrix(false);

	for(int i=0; i<4; i++ )
	{
		for(int j=0; j<4; j++ )
		{
			xResult.m_afElts[i][j] = 0.0f;

			for(int k=0; k<4; k++ )
			{
				xResult.m_afElts[i][j] += m_afElts[i][k] * xMat.m_afElts[k][j]; 
			}
		}
	}

	return xResult;
}

/*!
 * \brief  Performs a 3x3 matrix multiply ( this * xMat ). The fourth row/column returned comes from *this matrix
 * \param  xMat Matrix to multiply by
 * \return Result of the multiply operation
 */
Matrix Matrix::Multiply3x3( const Matrix &xMat ) const
{
	Matrix xResult = *this;

	for(int i=0; i<3; i++ )
	{
		for(int j=0; j<3; j++ )
		{
			xResult.m_afElts[i][j] = 0.0f;

			for(int k=0; k<3; k++ )
			{
				xResult.m_afElts[i][j] += m_afElts[i][k] * xMat.m_afElts[k][j]; 
			}
		}
	}

	return xResult;
}



/*!
 * \brief  Transforms (multiplies) a vector by a 4x4 matrix
 * \param  xIn Vector to transform
 * \return Result of the multiply operation
 */
Vector Matrix::PreTransform   ( const Vector &xIn ) const
{
	return Vector((xIn.m_fX * m_afElts[0][0] + xIn.m_fY * m_afElts[1][0] + xIn.m_fZ * m_afElts[2][0] + xIn.m_fW * m_afElts[3][0]),
		(xIn.m_fX * m_afElts[0][1] + xIn.m_fY * m_afElts[1][1] + xIn.m_fZ * m_afElts[2][1] + xIn.m_fW * m_afElts[3][1]),
		(xIn.m_fX * m_afElts[0][2] + xIn.m_fY * m_afElts[1][2] + xIn.m_fZ * m_afElts[2][2] + xIn.m_fW * m_afElts[3][2]),
		(xIn.m_fX * m_afElts[0][3] + xIn.m_fY * m_afElts[1][3] + xIn.m_fZ * m_afElts[2][3] + xIn.m_fW * m_afElts[3][3]));
}

/*!
 * \brief  Transforms (multiplies) a vector by a 3x3 matrix
 * \param  xIn Vector to transform
 * \return Result of the multiply operation
 */
Vector Matrix::PreTransform3x3( const Vector &xIn ) const
{
	return Vector((xIn.m_fX * m_afElts[0][0] + xIn.m_fY * m_afElts[1][0] + xIn.m_fZ * m_afElts[2][0]),
		(xIn.m_fX * m_afElts[0][1] + xIn.m_fY * m_afElts[1][1] + xIn.m_fZ * m_afElts[2][1]),
		(xIn.m_fX * m_afElts[0][2] + xIn.m_fY * m_afElts[1][2] + xIn.m_fZ * m_afElts[2][2]),
		xIn.m_fW);

}

/*!
 * \brief  Transforms (multiplies) a vector by a 3x4 matrix
 * \param  xIn Vector to transform
 * \return Result of the multiply operation
 */
Vector Matrix::PreTransform3x4( const Vector &xIn ) const
{
	return Vector((xIn.m_fX * m_afElts[0][0] + xIn.m_fY * m_afElts[1][0] + xIn.m_fZ * m_afElts[2][0] + xIn.m_fW * m_afElts[3][0]),
		(xIn.m_fX * m_afElts[0][1] + xIn.m_fY * m_afElts[1][1] + xIn.m_fZ * m_afElts[2][1] + xIn.m_fW * m_afElts[3][1]),
		(xIn.m_fX * m_afElts[0][2] + xIn.m_fY * m_afElts[1][2] + xIn.m_fZ * m_afElts[2][2] + xIn.m_fW * m_afElts[3][2]),
		xIn.m_fW);
}

/*!
 * \brief  Transforms (multiplies) a vector by the transpose of a 3x3 matrix
 * \param  xIn Vector to transform
 * \return Result of the multiply operation
 */
Vector Matrix::PostTransform3x3(const Vector &xIn ) const
{
	return Vector((xIn.m_fX * m_afElts[0][0] + xIn.m_fY * m_afElts[0][1] + xIn.m_fZ * m_afElts[0][2]),
		(xIn.m_fX * m_afElts[1][0] + xIn.m_fY * m_afElts[1][1] + xIn.m_fZ * m_afElts[1][2]),
		(xIn.m_fX * m_afElts[2][0] + xIn.m_fY * m_afElts[2][1] + xIn.m_fZ * m_afElts[2][2]),
		xIn.m_fW);
}

/*! Projection/Camera matrix creation */

/*!
 * \brief  Sets a perspective projection matrix (left-handed)
 * \param  fFoV Vertical FoV in radians
 * \param  fAspect Aspect ratio (Width/Height)
 * \param  fZNear Distance to near clip plane
 * \param  fZFar Distance to far clip plane
 */
void Matrix::SetPerspectiveMatrix( const float fFoV, const float fAspect, const float fZNear, const float fZFar)
{
	SetIdentity();

	float fYScale = (Math::Cos(fFoV * 0.5f)/Math::Sin(fFoV * 0.5f));
	float fXScale = fAspect * fYScale;
	float fZScale = fZFar / (fZFar - fZNear); 

	m_afElts[0][0] =  fXScale;
	m_afElts[1][1] =  fYScale;
	m_afElts[2][2] =  fZScale;
	m_afElts[3][2] = -fZNear * fZScale;
	m_afElts[2][3] =  1.0f;
	m_afElts[3][3] =  0.0f;
}

/*!
 * \brief  Sets an orthographic projection matrix (left-handed)
 * \param  fWidth Width of the view volume
 * \param  fHeight Height of the view volume
 * \param  fZNear Distance to near clip plane
 * \param  fZFar Distance to far clip plane
 */
void Matrix::SetOrthoMatrix( const float fWidth, const float fHeight, const float fZNear, const float fZFar)
{
	SetIdentity();

	m_afElts[0][0] =  2/fWidth;
	m_afElts[1][1] =  2/fHeight;
	m_afElts[2][2] =  1.0f/(fZFar-fZNear);
	m_afElts[3][2] =  fZNear/(fZNear-fZFar);
}

/*!
 * \brief  Sets a look-at matrix (left-handed)
 * \param  xFrom Camera position
 * \param  xAt Target position
 * \param  xUp Up vector (must be unit length)
 */
void Matrix::SetLookAtMatrix( const Vector &xFrom, const Vector &xAt, const Vector &xUp)
{
	Vector xAxisZ = (xAt - xFrom);
	xAxisZ.Normalize();
	Vector xAxisX = xUp.CrossProduct(xAxisZ);
	Vector xAxisY = xAxisZ.CrossProduct(xAxisX);

	(*this)[0] = Vector ( xAxisX.m_fX, xAxisY.m_fX, xAxisZ.m_fX, 0.0f);
	(*this)[1] = Vector ( xAxisX.m_fY, xAxisY.m_fY, xAxisZ.m_fY, 0.0f);
	(*this)[2] = Vector ( xAxisX.m_fZ, xAxisY.m_fZ, xAxisZ.m_fZ, 0.0f);
	(*this)[3] = Vector ( -xAxisX.DotProduct(xFrom), -xAxisY.DotProduct(xFrom), -xAxisZ.DotProduct(xFrom), 1.0f );
}
