/*! \file
 * Copyright (c) 2008
 * Project : Base
 * File    : Matrix Header (Independent)
 * Author  : Dave Reed
 */

/*-Include-Once--------------------------------------------------------------*/

#if !defined(__BASE_MATH_MATRIX_H__)
#define __BASE_MATH_MATRIX_H__ /*!< \brief Include once define. */

/*-Classes-------------------------------------------------------------------*/

/* Forward class declarations */
class Vector;

/*!
 * \brief The Matrix class - 4x4 Matrices
 */
class Matrix
{
public:

	/*! \name Constructors */
	Matrix( void );
	Matrix( const bool bSetIdentity );
	Matrix( const Matrix &xMtx );

	/*! \name Assignment operators */
	Matrix& operator = ( const Matrix &xV );

	/*! \name Transformation matrix creation */
	void SetIdentity   ( void );
	void SetTranslation( const Vector &xV );
	void SetScale      ( const float fX, const float fY, const float fZ );
	void SetRotationX  ( const float fRadians );
	void SetRotationY  ( const float fRadians );
	void SetRotationZ  ( const float fRadians );
	void SetRotation   ( const Vector xAxis, const float fRadians );

	/*! \name Transpose and Inverse */
	void  Transpose( void );
	void  Transpose3x3( void );
	void  InverseUnscaled( void );
	bool  Inverse( void );
	float Determinant( void ) const;
	float Determinant3x3( void ) const;

	/*! \name Decomposition */
	Vector GetRightVector( void ) const;
	Vector GetUpVector( void ) const;
	Vector GetForwardVector( void ) const;


	void   SetForwardVector( Vector &v );
	void   SetUpVector( Vector &v );
	void   SetRightVector( Vector &v );

	Vector GetTranslationVector( void ) const;
	void   SetTranslationVector( const Vector xV );

	/*! \name Indexing operator */
	Vector &Matrix::operator [] ( const int iElement ) const; // Returns matrix rows as Vectors. (Beware of W components != 1.0f)

	/*! \name Matrix x Matrix multiplication */
	Matrix operator * ( const Matrix &xMat ) const; // Matrix multiplication, on the full 4x4
	Matrix Multiply   ( const Matrix &xMat ) const; // Same as the operator *
	Matrix Multiply3x3( const Matrix &xMat ) const; // Matrix multiplication on the top left 3x3 submatrix only

	/*! \name Vector x Matrix multiplication */
	Vector PreTransform    ( const Vector &xIn ) const; // xIn * this, using whole matrix (can include projections)
	Vector PreTransform3x3 ( const Vector &xIn ) const; // xIn * this, using top left 3x3 of matrix (rot/scale only)
	Vector PreTransform3x4 ( const Vector &xIn ) const; // xIn * this, using top left 3x4 of matrix (rot/scale/translate)
	Vector PostTransform3x3( const Vector &xIn ) const; // this * xIn, using top left 3x3. (Multiply by transpose)

	/*! \name Projection/Camera matrix creation */
	void SetPerspectiveMatrix( const float fFoV, const float fAspect, const float fZNear, const float fZFar );
	void SetOrthoMatrix      ( const float fWidth, const float fHeight, const float fZNear, const float fZFar );
	void SetLookAtMatrix     ( const Vector &xFrom, const Vector &xAt, const Vector &xUp );

	/*! \name Private functions */
	void GetSubMatrix( Matrix &xDst, int iRowSkip, int iColSkip ) const; // Used by determinant/inverse functions

public:

	/*! \name Data */
	float m_afElts[4][4]; // Matrix elements, in [row][col] order
};


#endif // __BASE_MATH_MATRIX_H__
