/*! \file
 * Copyright (c) 2008
 * Project : Base
 * File    : Render Source (Independent)
 * Author  : Dave Reed
 */

/*-Project-Includes----------------------------------------------------------*/

#include "Main.h"
#include "Base/Render.h"
#include "Base/Blitter.h"
#include "Base/Shaders.h"
#include "Platform_Render.h"


/*-Class-Static-Variables----------------------------------------------------*/

RenderConfig Render::s_xCurrentConfig;	
float		 Render::s_f2DVirtualScreenWidth;
float		 Render::s_f2DVirtualScreenHeight;

/*-Class-Public-Functions----------------------------------------------------*/
/*!
 * \brief  Initialise graphics (e.g. create D3D device+window)
 * \param  Render Configuration
 * \return true if successful
 */
bool Render::Initialize(RenderConfig &xConfig)
{
	s_xCurrentConfig = xConfig;
	return Platform_Render::Initialize();
}

/*!
 * \brief  Shutdown graphics
 */
void Render::Shutdown(void)
{
	Platform_Render::Shutdown();
}

/*!
 * \brief  Reset graphics (if window resized, focus lost, or config changed)
 */
void Render::Reset(void)
{
	Platform_Render::Reset();
}

/*!
 * \brief  Does the system want the process to exit? (user hit the 'close' button on the graphics window)
 */
bool Render::IsExitRequested(void)
{
	return Platform_Render::IsExitRequested();
}

/*!
 * \brief  Present the current frame
 */
void Render::PresentFrame(void)
{
	Platform_Render::PresentFrame();
}

/*!
 * \brief  Get the screen width
 * \return Screen width in pixels
 */
int Render::GetScreenWidth(void)
{
	return Platform_Render::GetScreenWidth();
}

/*!
 * \brief  Get the screen height
 * \return Screen height in pixels
 */
int Render::GetScreenHeight(void)
{
	return Platform_Render::GetScreenHeight();
}

/*!
* \brief  Get the current viewport width
* \return Viewport width in pixels
*/
int Render::GetViewportWidth(void)
{
	return Platform_Render::GetViewportWidth();
}

/*!
* \brief  Get the current viewport height
* \return Viewport height in pixels
*/
int Render::GetViewportHeight(void)
{
	return Platform_Render::GetViewportHeight();
}

/*!
 * \brief  Is rendering paused (window minimized)
 * \return True if rendering shouldn't be done
 */
bool Render::IsRenderingPaused(void)
{
	return Platform_Render::IsRenderingPaused();
}

/*!
 * \brief  Is rendering to an in-focus window (if not, it's low priority)
 * \return True if render window is in focus
 */
bool Render::IsRenderingFocused(void)
{
	return Platform_Render::IsRenderingPaused();
}

/*!
* \brief  Flush any queued rendering operations, as we're about to end the frame/change camera/change viewport
*/
void Render::Flush(void)
{
	Blitter2D::Flush();
	Blitter3D::Flush();
}



void Render::ClearViewport(uint32 uCol, bool bColor, bool bDepth, bool bStencil)
{
	Platform_Render::ClearViewport(uCol, bColor, bDepth, bStencil);
}


float Render::GetAspectRatio()
{
	return Platform_Render::GetAspectRatio();
}


void Render::Reset2DScaling()
{
	// Resets the scaling to pixels
	Vector xScales;
	s_f2DVirtualScreenWidth  = (float)Render::GetViewportWidth();
	s_f2DVirtualScreenHeight = (float)Render::GetViewportHeight();
	xScales.m_fX =  2.0f/((float)Render::GetViewportWidth());
	xScales.m_fY = -2.0f/((float)Render::GetViewportHeight());
	xScales.m_fZ = -1.0f + (-2.0f/((float)Render::GetViewportWidth()));
	xScales.m_fW =  1.0f - (-2.0f/((float)Render::GetViewportHeight()));
	Shaders::SetSharedParamVector4(SHADER_PARAM_PIXELSTOVIEWPORT, xScales);
}


void Render::Set2DVirtualCoords(float fHeight)
{
	Vector xScales;
	float fAspect = GetAspectRatio();
	float fWidth = fHeight / fAspect;
	s_f2DVirtualScreenWidth  = fWidth;
	s_f2DVirtualScreenHeight = fHeight;
	xScales.m_fX =  2.0f/((float)fWidth);
	xScales.m_fY = -2.0f/((float)fHeight);
	xScales.m_fZ = -1.0f + (-2.0f/((float)fWidth));
	xScales.m_fW =  1.0f - (-2.0f/((float)fHeight));
	Shaders::SetSharedParamVector4(SHADER_PARAM_PIXELSTOVIEWPORT, xScales);
}

void Render::Reset2DTransform()
{
	// Resets to an identity xform
	Matrix xTmp;
	xTmp.SetIdentity();
	Shaders::SetSharedParamMatrix44(SHADER_PARAM_2DXFORM, xTmp);
}

void Render::Set2DTransform(float fX, float fY, float fRotation, float fXScale, float fYScale)
{
	// Rotate/scale around 0,0, and then shift to fX, fY
	Matrix xFinal, xRot, xScale;
	xRot.SetRotationZ(-fRotation);
	xScale.SetScale(fXScale, fYScale, 1.0f);
	xFinal = xScale * xRot;
	Vector v = Vector(fX, fY, 0.0f);
	xFinal.SetTranslationVector(v);
	Shaders::SetSharedParamMatrix44(SHADER_PARAM_2DXFORM, xFinal);
}

float Render::Get2DScreenLeftX()
{
	return 0.0f;
}

float Render::Get2DScreenRightX()
{
	return s_f2DVirtualScreenWidth;
}

float Render::Get2DScreenCentreX()
{
	return s_f2DVirtualScreenWidth*0.5f;
}


float Render::Get2DScreenTopY()
{
	return 0.0f;
}

float Render::Get2DScreenBottomY()
{
	return s_f2DVirtualScreenHeight;
}

float Render::Get2DScreenMiddleY()
{
	return s_f2DVirtualScreenHeight / 2.0f;
}

void Render::SetWorldMatrix(Matrix& mtx)
{
	Platform_Render::SetWorldMatrix(mtx);
}