/*! \file
 * Copyright (c) 2008
 * Project : Fulcrum Engine
 * File    : Vector Source (Independent)
 * Author  : Dave Reed
 */

/*-Project-Includes----------------------------------------------------------*/

#include "Main.h"
#include "Base/Vector.h"
#include "Base/Math.h"
#include "Base/Matrix.h"

/*-Class-Public-Functions----------------------------------------------------*/


/* Constructors */

/*!
 * \brief  Default constructor (Inits components to 0,0,0,1)
 */
Vector::Vector( void )
{
	m_fX = m_fY = m_fZ = 0.0f;
	m_fW = 1.0f;
}

/*!
 * \brief  Constructor from 3 or 4 floats
 * \param  fX The X component of the vector.
 * \param  fY The Y component of the vector.
 * \param  fZ The Z component of the vector.
 * \param  fW The W component of the vector (Optional, default = 1.0f).
 */
Vector::Vector( const float fX, const float fY, const float fZ, const float fW )
{
	m_fX = fX;
	m_fY = fY;
	m_fZ = fZ;
	m_fW = fW;
}

/*!
 * \brief Copy constructor
 * \param  xV Vector to copy from
 */
Vector::Vector( const Vector &xV )
{
	*this = xV;
}

/* Assignment operators */

/*!
 * \brief  Assigment from another Vector
 * \param  xV Vector to copy from
 * \return A reference to this Vector
 */
Vector& Vector::operator = ( const Vector &xV )
{
	m_fX = xV.m_fX;
	m_fY = xV.m_fY;
	m_fZ = xV.m_fZ;
	m_fW = xV.m_fW;
	return *this;
}

/*!
 * \brief  Negation Operator
 * \return A reference to this Vector
 */
Vector Vector::operator - () const
{
	Vector xRet = Vector(-m_fX, -m_fY, -m_fZ, m_fW);
	return xRet;
}

/*!
 * \brief Indexing operator
 * \param iElement Element to get (0-3)
 * \return Selected element of the vector
 */
float &Vector::operator[] ( const int iElement ) 
{
	float  *pfVec = (float *)(this);
	return (float &)(pfVec[iElement]);
}

/*!
 * \brief  Addition assigment from another Vector
 * \param  xV Vector to add
 * \return A reference to this Vector
 */
Vector& Vector::operator += ( const Vector &xV )
{
	m_fX += xV.m_fX;
	m_fY += xV.m_fY;
	m_fZ += xV.m_fZ;
	return *this;
}

/*!
 * \brief  Subtraction assigment from another Vector
 * \param  xV Vector to subtract
 * \return A reference to this Vector
 */
Vector& Vector::operator -= ( const Vector &xV )
{
	m_fX -= xV.m_fX;
	m_fY -= xV.m_fY;
	m_fZ -= xV.m_fZ;
	return *this;
}

/*!
 * \brief  Multiplication assigment (multiply by a scalar)
 * \param  fValue Value to multiply by
 * \return A reference to this Vector
 */
Vector& Vector::operator *= ( float fValue )
{
	m_fX *= fValue;
	m_fY *= fValue;
	m_fZ *= fValue;
	return *this;
} 

/*!
 * \brief  Division assigment (divide by a scalar)
 * \param  fValue Value to multiply by
 * \return A reference to this Vector
 */
Vector& Vector::operator /= ( float fValue )
{
	m_fX /= fValue;
	m_fY /= fValue;
	m_fZ /= fValue;
	return *this;
}

/* Arithmetic operators */

/*!
 * \brief  Addition operator
 * \param  xV Vector to add
 * \return Result of the addition
 */
Vector Vector::operator + ( const Vector &xV ) const
{
	Vector xRet = Vector(m_fX + xV.m_fX,
		m_fY + xV.m_fY,
		m_fZ + xV.m_fZ,
		m_fW);
	return xRet;
}

/*!
 * \brief  Subtraction operator
 * \param  xV Vector to subtract
 * \return Result of the subtraction
 */
Vector Vector::operator - ( const Vector &xV ) const
{
	Vector xRet = Vector(m_fX - xV.m_fX,
		m_fY - xV.m_fY,
		m_fZ - xV.m_fZ,
		m_fW);
	return xRet;
}

/*!
 * \brief  Scalar multiplication operator
 * \param  fValue Scalar to multiply by
 * \return Result of the multiplication
 */
Vector Vector::operator * ( float fValue ) const
{
	Vector xRet;
	xRet.m_fX = m_fX * fValue;
	xRet.m_fY = m_fY * fValue;
	xRet.m_fZ = m_fZ * fValue;
	return xRet;
}

/*!
 * \brief  Scalar division operator
 * \param  fValue Scalar to divide by
 * \return Result of the division
 */
Vector Vector::operator / ( float fValue ) const
{
	Vector xRet;
	xRet.m_fX = m_fX / fValue;
	xRet.m_fY = m_fY / fValue;
	xRet.m_fZ = m_fZ / fValue;
	return xRet;
}

/*!
 * \brief  Multiplication by a 4x4 matrix [Same as Matrix::PreTransform()]
 * \param  xMtx Matrix to multiply (transform) by
 * \return Result of the multiplication
 */
Vector Vector::operator * ( Matrix &xMtx ) const
{
	return xMtx.PreTransform(*this);
}

/* Comparison Operators */

/*!
 * \brief  Equality operator
 * \param  xV Vector to compare with
 * \return true if the Vector's are exactly equal
 */
bool Vector::operator == ( const Vector &xV ) const
{
	return (m_fX == xV.m_fX && 
		m_fY == xV.m_fY && 
		m_fZ == xV.m_fZ && 
		m_fW == xV.m_fW );
}

/*!
 * \brief  Not Equal operator
 * \param  xV Vector to compare with
 * \return true if the Vector's are not exactly equal
 */
bool Vector::operator != ( const Vector &xV ) const
{
	return !(*this == xV);
}

/*!
 * \brief  Approximately Equal function
 * \param  xV Vector to compare with
 * \return true if the Vector's are within 0.00001f of each other
 */
bool Vector::ApproxEquals( const Vector &xV ) const
{
	float fThreshold = 0.00001f;

	return (( xV.m_fX > m_fX - fThreshold && xV.m_fX < m_fX + fThreshold ) &&
		( xV.m_fY > m_fY - fThreshold && xV.m_fY < m_fY + fThreshold ) &&
		( xV.m_fZ > m_fZ - fThreshold && xV.m_fZ < m_fZ + fThreshold ) &&
		( xV.m_fW > m_fW - fThreshold && xV.m_fW < m_fW + fThreshold ));
}

/* Dot and Cross products (only 3-dimensional, ignores W component) */

/*!
 * \brief  3-component (xyz) dot product. Ignores W components
 * \param  xV Vector to dot with
 * \return result of the dot product
 */
float Vector::DotProduct  ( const Vector& xV ) const
{
	return (m_fX * xV.m_fX +
		m_fY * xV.m_fY +
		m_fZ * xV.m_fZ);
}

/*!
 * \brief  3-component (xyz) cross product. Ignores W components
 * \param  xV Vector to cross with
 * \return result of the cross product
 */
Vector Vector::CrossProduct( const Vector& xV ) const
{
	Vector xOut;
	xOut.m_fX = m_fY * xV.m_fZ - m_fZ * xV.m_fY;
	xOut.m_fY = m_fZ * xV.m_fX - m_fX * xV.m_fZ;
	xOut.m_fZ = m_fX * xV.m_fY - m_fY * xV.m_fX;
	xOut.m_fW = 1.0f;
	return xOut;
}

/* Magnitude/Normalization */

/*!
 * \brief  Magnitude Squared function
 * \return The vector's magnitude squared
 */
float Vector::MagnitudeSquared( void ) const
{
	return (m_fX * m_fX +
		    m_fY * m_fY +
		    m_fZ * m_fZ);
}

/*!
 * \brief  Magnitude function
 * \return The vector's magnitude
 */
float Vector::Magnitude( void ) const
{
	return Math::Sqrt(MagnitudeSquared());
}

/*!
 * \brief  Normalize function
 * \return The vector's magnitude (before normalization!)
 */
float Vector::Normalize( void )
{
	float fMagnitude = Magnitude();

	if (fMagnitude!=0.0f)
	{
		m_fX /= fMagnitude;
		m_fY /= fMagnitude;
		m_fZ /= fMagnitude;
	}

	return fMagnitude;
}

/* Interpolation */

/*!
 * \brief  Linear interpolation function
 * \param  xTarget Destination vector
 * \param  fLerp Interpolation value (0.0f-1.0f)
 * \return The interpolated vector
 */
Vector Vector::Lerp( const Vector &xTarget, const float fLerp ) const
{
	float fLerpClamped = Math::Clamp(fLerp,0.0f,1.0f);
	float fOneMinusLerp = 1.0f - fLerpClamped;

	Vector xOutput = Vector( m_fX * fOneMinusLerp + xTarget.m_fX * fLerpClamped,
		m_fY * fOneMinusLerp + xTarget.m_fY * fLerpClamped,
		m_fZ * fOneMinusLerp + xTarget.m_fZ * fLerpClamped,
		m_fW * fOneMinusLerp + xTarget.m_fW * fLerpClamped);

	return xOutput;
}


/*!
* \brief  Catmull-Rom spline interpolation function
* \param  p0-p3 Control points
* \param  s Interpolation value - 0..1
*/
void Vector::CatmullRom( const Vector &p0, const Vector &p1, const Vector &p2, const Vector &p3, float s)
{
//	Q(s) = [(-s3 + 2s2 - s)p1 + (3s3 - 5s2 + 2)p2 + (-3s3 + 4s2 + s)p3 + (s3 - s2)p4] / 2

	float s2 = s*s;
	float s3 = s2*s;

	float Scale0 = 0.5f * (  -s3  + 2*s2 - s);
	float Scale1 = 0.5f * ( 3*s3  - 5*s2 + 2);
	float Scale2 = 0.5f * (-3*s3  + 4*s2 + s);
	float Scale3 = 0.5f * (   s3  -   s2    );

	*this = (p0 * Scale0) + (p1 * Scale1) + (p2 * Scale2) + (p3 * Scale3);
}